/************************************************************************
*                                                                       *
*                                 |\             ,                      *
*                      _           \\   '       ||                      *
*                     / \\  /'\\  / \\ \\  _-_ =||=                     *
*                    || || || || || || || ||    ||                      *
*                    || || || || || || || ||    ||                      *
*                    \\_-| \\,/   \\/  \\ \\,/  \\,                     *
*                     /  \                                              *
*                    '----`                                             *
*                                                                       *
*                Copyright (c) 1993/94 Jan van der Steen                *
*              Centre for Mathematics and Computer Science              *
*                      Amsterdam, the Netherlands                       *
*                                                                       *
*-----------------------------------------------------------------------*
*   You are not allowed to use this software in commercial software.    *
*   You are allowed to use this software in free software under the     *
*   condition that you maintain this integral message in the source.    *
*-----------------------------------------------------------------------*
* File    : %M%                                                         *
* Purpose : ___________________________________________________________ *
* Version : %I%                                                         *
* Modified: %G% %U%                                                     *
* Author  : Jan van der Steen (jansteen@cwi.nl)                         *
************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "godict.h"
#include "io.h"

static int	interactive = 0;
static int	html        = 0;	/* HTML or TEXT	*/
static int	verbose	    = 0;
static int	types	    = 0;	/* No types	*/
static int	langs	    = 0;	/* No languages	*/

static GODICT *	dict_search	(GODICT *d, char *p, FOUND *f);
static int	dict_print	(GODICT *d, int emit);
static void	dict_output	(GODICT *d, char *p, FOUND *f);
static void	dict_collect	(GODICT *d, char *p, FOUND *f);
static void	dict_report	(FOUND *f);
static void	dict_close	(GODICT *d);
static void	dict_store	(char **f, char *s, int lower);
static void	dict_img	(GODICT *d, char *s);
static int	dict_fmt	(char *label, char *s, int emit);


#define LANGENTRY(line,label)	(!strncmp(line,label,strlen(label)))

GODICT *
dict_load()
{
    FILE   * fp;
    GODICT * d	      = (GODICT *) 0;
    GODICT * dp	      = (GODICT *) 0;
    char   * dictname = getenv("INTERGO");
    int	     linenr   = 0;
    int	     record   = 0;
    char     line[MAXDICTLINE];

    /*
     * If the environment variable INTERGO is not set we take the default
     */
    if (dictname == NULL) dictname = DEFDICT;

    /*
     * Open the dictionary
     */
    if ((fp = fopen(dictname, "r")) == NULL) {
	msg(MSG_SYS_ERR, "Can't open \"%s\"\n", dictname);
	return d;
    }

    /*
     * Read the dictionary
     */
    while (fgets(line, MAXDICTLINE, fp) != NULL) {
	char *nl;

	/*
	 * Account line and move the \n (and possible \r)
	 */
	linenr++;
	if ((nl = strchr(line, '\n')) != NULL) *nl = 0;
	if ((nl = strchr(line, '\r')) != NULL) *nl = 0;

	/*
	 * Comment or empty lines?
	 */
	if (*line == 0 || *line == COMMENT) continue;

	/*
	 * Bad input lines?
	 */
	if (strlen(line) < strlen(RD_CD)) {
	    msg(MSG_SYS_ERR, "Line %d, Bad format: \"%s\"\n",
		    linenr,
		    line);
	    continue;
	}

	if (strncmp(line, RD_CD, strlen(RD_CD)) == 0) {
	    /*
	     * Deal with new dictionary entry after closing previous one
	     */
	    if (d) dict_close(d);

	    /*
	     * Allocate and insert new dictionary entry
	     */
	    dp = (GODICT *) calloc(1, sizeof(GODICT));
	    if (dp == NULL) {
		msg(MSG_SYS_ERR, "Out of memory in dict_load()\n");
		fclose(fp);
		return dp;
	    }
	    dp->dct_next = d;
	    d = dp;
	    switch(*(line+strlen(RD_CD))) {
		case 'n': d->dct_type = CD_NAME; break;
		case 'c': d->dct_type = CD_CHAM; break;
		case 't': d->dct_type = CD_TECH; break;
		case 'p': d->dct_type = CD_POLI; break;
		case 'd': d->dct_type = CD_DIGI; break;
		default : d->dct_type = CD_MISC; break;
	    }
	    record++;

	    /*
	     * Report...
	     */
	    if (verbose && (record % 10) == 0) msg(MSG_USR_ERR, "[%3d]", record);

	    continue;
	}

	/*
	 * Check for language definition
	 */
	if (d == NULL) {
	    msg(MSG_SYS_ERR, "Line %d, Skipped: \"%s\"\n",
		    linenr,
		    line);
	}
	if LANGENTRY(line,RD_JP) { dict_store(&(d->dct_jp), line+strlen(RD_JP), 1); continue; }
	if LANGENTRY(line,RD_CH) { dict_store(&(d->dct_ch), line+strlen(RD_CH), 1); continue; }
	if LANGENTRY(line,RD_RK) { dict_store(&(d->dct_rk), line+strlen(RD_RK), 1); continue; }
	if LANGENTRY(line,RD_GB) { dict_store(&(d->dct_gb), line+strlen(RD_GB), 1); continue; }
	if LANGENTRY(line,RD_NL) { dict_store(&(d->dct_nl), line+strlen(RD_NL), 1); continue; }
	if LANGENTRY(line,RD_GE) { dict_store(&(d->dct_ge), line+strlen(RD_GE), 1); continue; }
	if LANGENTRY(line,RD_FR) { dict_store(&(d->dct_fr), line+strlen(RD_FR), 1); continue; }
	if LANGENTRY(line,RD_SV) { dict_store(&(d->dct_sv), line+strlen(RD_SV), 1); continue; }
	if LANGENTRY(line,RD_IT) { dict_store(&(d->dct_it), line+strlen(RD_IT), 1); continue; }
	if LANGENTRY(line,RD_EO) { dict_store(&(d->dct_eo), line+strlen(RD_EO), 1); continue; }
	if LANGENTRY(line,RD_DG) { dict_store(&(d->dct_dg), line+strlen(RD_DG), 0); continue; }
	if LANGENTRY(line,RD_CP) { dict_store(&(d->dct_cp), line+strlen(RD_CP), 0); continue; }
	if LANGENTRY(line,RD_RF) { dict_store(&(d->dct_rf), line+strlen(RD_RF), 0); continue; }
	if LANGENTRY(line,RD_GF) { dict_img  (d, line+strlen(RD_GF)); continue; }
	msg(MSG_SYS_ERR, "Line %d, Unsupported language: \"%s\"\n",
		linenr,
		line);
    }

    /*
     * Close last dictionary entry
     */
    if (d != NULL) dict_close(d);

    /*
     * Report...
     */
    if (verbose) {
	if (record % 10 != 0) msg(MSG_USR_ERR, "[%3d]", record);
	msg(MSG_USR_ERR, "\n");
    }

    fclose(fp);
    return d;
}

static void
dict_store(f, s, lower)
char **f;	/* Field */
char  *s;	/* Data  */
int    lower;	/* Convert to lowercase */
{
    int  more   = (*f != NULL);
    long needed = strlen(s) + 1;

    /*
     * Don't store empty fields
     */
    if (!*s) return;

    if (more) {
	needed += strlen(*f) + 1;
	*f = realloc(*f, needed);
    }
    else {
	*f = malloc(needed);
    }

    if (*f == NULL) {
	msg(MSG_SYS_ERR, "Out of memory in dict_store()\n");
	exit(1);
    }
    if (more) {
	strcat(*f, "\n");
	strcat(*f, (lower) ? lstr(s) : s);
    }
    else {
	strcpy(*f, (lower) ? lstr(s) : s);
    }
}

static void
dict_img(d, s)
/*
 * GF=seki.gif WD=120 HD=90
 */
GODICT *d;	/* Record */
char   *s;	/* Data   */
{
    char *key;

    /*
     * Don't store empty fields
     */
    if (!*s) return;

    if ((key = strchr(s, ' '))       == NULL) goto syntax_error;
    *key = 0;
    if ((key = strstr(key+1, "WD=")) == NULL) goto syntax_error;
    d->dct_dimen[0] = atoi(key+3);
    if ((key = strchr(key+1, ' '))   == NULL) goto syntax_error;
    if ((key = strstr(key+1, "HD=")) == NULL) goto syntax_error;
    d->dct_dimen[1] = atoi(key+3);
    if ((d->dct_gf = malloc(strlen(s) + 1)) == NULL) {
	msg(MSG_SYS_ERR, "Out of memory in dict_img()\n");
	exit(1);
    }
    strcpy(d->dct_gf, s);
    return;
syntax_error:
    fprintf(stderr, "Syntax error in image directive\n");
}

static void
dict_close(d)
/*
 * Close a dictionary entry
 */
GODICT *d;
{
    if (!d) {
	msg(MSG_SYS_ERR, "Null pointer in dict_close()\n");
	exit(1);
    }
#if 0
    /*
     * If a name entry is empty it's assumed to be
     * the same as the Japanese transcription
     * (which might also be empty of course)
     */
    if (d->dct_type == CD_NAME) {
	if (!d->dct_gb) d->dct_gb = d->dct_jp;
	if (!d->dct_nl) d->dct_nl = d->dct_jp;
	if (!d->dct_ge) d->dct_ge = d->dct_jp;
    }
#endif
}

#define SETL(f,l)		((f)->fnd_lang |= (l))
#define WORDINLANG(df,l,w)	((df) && (langs & (l)) && strstr((df),(w)))

static GODICT *
dict_search(d, s, f)
/*
 * Start querying the database for string s at entry point d
 */
GODICT	*d;
char	*s;
FOUND	*f;
{
    int found = 0;

    while (d) {
      if (types & d->dct_type) {
	if (WORDINLANG(d->dct_jp,LANG_JP,s)) { SETL(f,LANG_JP); found++; }
	if (WORDINLANG(d->dct_ch,LANG_CH,s)) { SETL(f,LANG_CH); found++; }
	if (WORDINLANG(d->dct_rk,LANG_RK,s)) { SETL(f,LANG_RK); found++; }
	if (WORDINLANG(d->dct_gb,LANG_GB,s)) { SETL(f,LANG_GB); found++; }
	if (WORDINLANG(d->dct_nl,LANG_NL,s)) { SETL(f,LANG_NL); found++; }
	if (WORDINLANG(d->dct_ge,LANG_GE,s)) { SETL(f,LANG_GE); found++; }
	if (WORDINLANG(d->dct_fr,LANG_FR,s)) { SETL(f,LANG_FR); found++; }
	if (WORDINLANG(d->dct_sv,LANG_SV,s)) { SETL(f,LANG_SV); found++; }
	if (WORDINLANG(d->dct_it,LANG_IT,s)) { SETL(f,LANG_IT); found++; }
	if (WORDINLANG(d->dct_eo,LANG_EO,s)) { SETL(f,LANG_EO); found++; }
	if (WORDINLANG(d->dct_dg,LANG_DG,s)) { SETL(f,LANG_DG); found++; }
	if (WORDINLANG(d->dct_cp,LANG_CP,s)) { SETL(f,LANG_CP); found++; }
	if (WORDINLANG(d->dct_rf,LANG_RF,s)) { SETL(f,LANG_RF); found++; }
	if (WORDINLANG(d->dct_gf,LANG_GF,s)) { SETL(f,LANG_GF); found++; }
	if (found) return d;
      }
      d = d->dct_next;
    }
    return d;
}

static int
dict_fmt(label, s, emit)
/*
 * Output a field value which might possibly contain newlines.
 * We output the label for each line.
 * If emit=0, we just return the number of bytes which *would* be emitted.
 */
char *label;
char *s;
int   emit;
{
    int n = 0;

    if (emit) {
	char def[MAXDICTLINE];
	char *t = def;

	while (*s) {
	    msg(MSG_USR_OUT, "%s", label);
	    while (*s) {
		*t++ = *s;
		if (*s++ == '\n') {
		    *t = 0;
		    msg(MSG_USR_OUT, "%s", def);
		    t = def;
		    break;
		}
	    }
	}
	*t = 0;
	msg(MSG_USR_OUT, "%s\n", def);
    }
    else {
	while (*s) {
	    n += (int) strlen(label);
	    while (*s) {
		n++;
		if (*s++ == '\n') break;
	    }
	}
	n++;
    }
    return n;
}

#define LANGDEF(df,l)	((langs & (l)) && (df) && *(df))

static int
dict_print(d, emit)
/*
 * Print a dictionary entry
 */
GODICT *d;
int	emit;
{
    int n = 0;

    if (d == NULL) {
	msg(MSG_SYS_ERR, "Null pointer in dict_print()\n");
	exit(1);
    }

#if SERVER
    switch(d->dct_type) {
	case CD_CHAM: n += dict_fmt(LB_CD, MSG_CHAM, emit); break;
	case CD_TECH: n += dict_fmt(LB_CD, MSG_TECH, emit); break;
	case CD_NAME: n += dict_fmt(LB_CD, MSG_NAME, emit); break;
	case CD_POLI: n += dict_fmt(LB_CD, MSG_POLI, emit); break;
	case CD_DIGI: n += dict_fmt(LB_CD, MSG_DIGI, emit); break;
	default	    : n += dict_fmt(LB_CD, MSG_MISC, emit); break;
    }
#else
    if (d->dct_spec && *d->dct_spec) n += dict_fmt(LB_CD, d->dct_spec, emit);
#endif

    if (LANGDEF(d->dct_jp,LANG_JP)) n += dict_fmt(LB_JP, d->dct_jp, emit);
    if (LANGDEF(d->dct_ch,LANG_CH)) n += dict_fmt(LB_CH, d->dct_ch, emit);
    if (LANGDEF(d->dct_rk,LANG_RK)) n += dict_fmt(LB_RK, d->dct_rk, emit);
    if (LANGDEF(d->dct_gb,LANG_GB)) n += dict_fmt(LB_GB, d->dct_gb, emit);
    if (LANGDEF(d->dct_nl,LANG_NL)) n += dict_fmt(LB_NL, d->dct_nl, emit);
    if (LANGDEF(d->dct_ge,LANG_GE)) n += dict_fmt(LB_GE, d->dct_ge, emit);
    if (LANGDEF(d->dct_fr,LANG_FR)) n += dict_fmt(LB_FR, d->dct_fr, emit);
    if (LANGDEF(d->dct_sv,LANG_SV)) n += dict_fmt(LB_SV, d->dct_sv, emit);
    if (LANGDEF(d->dct_it,LANG_IT)) n += dict_fmt(LB_IT, d->dct_it, emit);
    if (LANGDEF(d->dct_eo,LANG_EO)) n += dict_fmt(LB_EO, d->dct_eo, emit);
    if (LANGDEF(d->dct_dg,LANG_DG)) n += dict_fmt(LB_DG, d->dct_dg, emit);
    if (LANGDEF(d->dct_cp,LANG_CP)) n += dict_fmt(LB_CP, d->dct_cp, emit);
    if (LANGDEF(d->dct_rf,LANG_RF)) n += dict_fmt(LB_RF, d->dct_rf, emit);
    if (LANGDEF(d->dct_gf,LANG_GF)) n += dict_fmt(LB_GF, d->dct_gf, emit);

    return n;
}

static void
dict_2html(GODICT *d, char *gifpath)
/*
 * Convert a dictionary entry to HTML
 */
{
    if (d == NULL) {
	msg(MSG_SYS_ERR, "Null pointer in dict_2html()\n");
	exit(1);
    }
#if 0
    switch(d->dct_type) {
	case CD_CHAM: n += dict_fmt(LB_CD, MSG_CHAM, emit); break;
	case CD_TECH: n += dict_fmt(LB_CD, MSG_TECH, emit); break;
	case CD_NAME: n += dict_fmt(LB_CD, MSG_NAME, emit); break;
	case CD_POLI: n += dict_fmt(LB_CD, MSG_POLI, emit); break;
	case CD_DIGI: n += dict_fmt(LB_CD, MSG_DIGI, emit); break;
	default	    : n += dict_fmt(LB_CD, MSG_MISC, emit); break;
    }
#endif
    if (LANGDEF(d->dct_jp,LANG_JP)) {
	fprintf(stdout, "<tr>\n");
	fprintf(stdout, "<th width=\"20%%\" class=l>Japanese</th>\n");
	fprintf(stdout, "<td width=\"80%%\"><strong>%s</strong></td>\n", d->dct_jp);
	fprintf(stdout, "</tr>\n");
    }
    if (LANGDEF(d->dct_gf,LANG_GF)) {
	fprintf(stdout, "<tr><td width=\"100%%\" align=center colspan=2>\n");
	fprintf(stdout, "<table><tr><td valign=bottom>");
	fprintf(stdout, "<table border=2><tr><td>");
	fprintf(stdout, "<img src=\"%s/%s\" width=%d height=%d>",
		gifpath,
		d->dct_gf,
		d->dct_dimen[0],
		d->dct_dimen[1]);
	fprintf(stdout, "</td></tr></table>\n");
	fprintf(stdout, "</td><td valign=bottom>\n");
	if (LANGDEF(d->dct_cp,LANG_CP)) {
	    fprintf(stdout, "<br>\n");
	    fprintf(stdout, "<em>\n");
	    /* fprintf(stdout, "<u>Diagram</u>:\n"); */
	    fprintf(stdout, "%s\n", d->dct_cp);
	    fprintf(stdout, "</em>\n");
	}
	fprintf(stdout, "</td></tr></table>\n");
	fprintf(stdout, "</td></tr>\n");
    } else {
	if (LANGDEF(d->dct_dg,LANG_DG)) {
	    fprintf(stdout, "<tr>\n");
	    fprintf(stdout, "<td colspan=2><small><pre>\n");
	    fprintf(stdout, "%s\n\n", d->dct_dg);
	    fprintf(stdout, "</pre></small></td>\n");
	    fprintf(stdout, "</tr>\n");
	}
	if (LANGDEF(d->dct_cp,LANG_CP))
	    fprintf(stdout, "<tr><td colspan=2>%s</td></tr>\n", d->dct_cp);
    }
    if (LANGDEF(d->dct_ch,LANG_CH)) {
	fprintf(stdout, "<tr>\n");
	fprintf(stdout, "<th width=\"20%%\" class=l>Chinese</th>\n");
	fprintf(stdout, "<td width=\"80%%\">%s</td>\n", d->dct_ch);
	fprintf(stdout, "</tr>\n");
    }
    if (LANGDEF(d->dct_rk,LANG_RK)) {
	fprintf(stdout, "<tr>\n");
	fprintf(stdout, "<th width=\"20%%\" class=l>Korean</th>\n");
	fprintf(stdout, "<td width=\"80%%\">%s</td>\n", d->dct_rk);
	fprintf(stdout, "</tr>\n");
    }
    if (LANGDEF(d->dct_gb,LANG_GB)) {
	fprintf(stdout, "<tr>\n");
	fprintf(stdout, "<th width=\"20%%\" class=l>English</th>\n");
	fprintf(stdout, "<td width=\"80%%\">%s</td>\n", d->dct_gb);
	fprintf(stdout, "</tr>\n");
    }
    if (LANGDEF(d->dct_nl,LANG_NL))
	fprintf(stdout, "<tr><th class=l>Dutch</th><td>%s</td></tr>\n", d->dct_nl);
    if (LANGDEF(d->dct_ge,LANG_GE))
	fprintf(stdout, "<tr><th class=l>German</th><td>%s</td></tr>\n", d->dct_ge);
    if (LANGDEF(d->dct_fr,LANG_FR))
	fprintf(stdout, "<tr><th class=l>French</th><td>%s</td></tr>\n", d->dct_fr);
    if (LANGDEF(d->dct_sv,LANG_SV))
	fprintf(stdout, "<tr><th class=l>Swedisch</th><td>%s</td></tr>\n", d->dct_sv);
    if (LANGDEF(d->dct_it,LANG_IT))
	fprintf(stdout, "<tr><th class=l>Italian</th><td>%s</td></tr>\n", d->dct_it);
    if (LANGDEF(d->dct_eo,LANG_EO))
	fprintf(stdout, "<tr><th class=l>Esperanto</th><td>%s</td></tr>\n", d->dct_eo);
    if (LANGDEF(d->dct_rf,LANG_RF)) {
	char *reference = "Reference";
	if (d->dct_jp && *d->dct_jp) reference = d->dct_jp; else
	if (d->dct_gb && *d->dct_gb) reference = d->dct_gb;
	fprintf(stdout, "<tr><th class=l>WWW</th><td><a href=\"%s\">%s</a></td></tr>\n",
		d->dct_rf, reference);
    }
}

static void
dict_report(f)
FOUND *f;
{
    if (f->fnd_records) {
	msg(MSG_USR_OUT,
	    "[%d records (%ld bytes) matched in ",
	    f->fnd_records, f->fnd_bytes);
	if (f->fnd_lang & LANG_JP) msg(MSG_USR_OUT, "[JP]");
	if (f->fnd_lang & LANG_CH) msg(MSG_USR_OUT, "[CH]");
	if (f->fnd_lang & LANG_GB) msg(MSG_USR_OUT, "[GB]");
	if (f->fnd_lang & LANG_RK) msg(MSG_USR_OUT, "[RK]");
	if (f->fnd_lang & LANG_NL) msg(MSG_USR_OUT, "[NL]");
	if (f->fnd_lang & LANG_GE) msg(MSG_USR_OUT, "[GE]");
	if (f->fnd_lang & LANG_FR) msg(MSG_USR_OUT, "[FR]");
	if (f->fnd_lang & LANG_SV) msg(MSG_USR_OUT, "[SV]");
	if (f->fnd_lang & LANG_IT) msg(MSG_USR_OUT, "[IT]");
	if (f->fnd_lang & LANG_DG) msg(MSG_USR_OUT, "[DG]");
	if (f->fnd_lang & LANG_CP) msg(MSG_USR_OUT, "[CP]");
	if (f->fnd_lang & LANG_RF) msg(MSG_USR_OUT, "[RF]");
	if (f->fnd_lang & LANG_GF) msg(MSG_USR_OUT, "[GF]");
	msg(MSG_USR_OUT, "]\n");
    } else msg(MSG_USR_OUT, "[No records matched]\n");
}

static void
dict_output(d, p, f)
/*
 * Output all dictionary entries which contain pattern "p"
 */
GODICT *d;		/* Dictionary		*/
char   *p;		/* Search pattern	*/
FOUND  *f;		/* What did we find?	*/
{
    char   *gif = getenv("INTERGO_GIF");
    GODICT *dp;
    int     next = 0;

    if (html) {
	if (!gif) gif = "http://www.cwi.nl/~jansteen/intergo/diagrams";
#if 0
	fprintf(stdout, "<HTML>\n");
	fprintf(stdout, "<HEAD>\n");
	fprintf(stdout, "<TITLE>\n");
	fprintf(stdout, "Intergo search for %s\n", p);
	fprintf(stdout, "</TITLE>\n");
	fprintf(stdout, "</HEAD>\n");
	fprintf(stdout, "<BODY text=\"#000000\"\n");
	fprintf(stdout, "   bgcolor=\"#dddddd\"\n");
	fprintf(stdout, "      link=\"#8a1818\"\n");
	fprintf(stdout, "     alink=\"#cf2424\"\n");
	fprintf(stdout, "     vlink=\"#cf2424\">\n");
	fprintf(stdout, "<h2>\n");
	fprintf(stdout, "Intergo search for %s\n", p);
	fprintf(stdout, "</h2>\n");
#endif
    }
    for (dp=dict_search(d, p, f); dp; dp=dict_search(dp->dct_next, p, f)) {
	if (next++) msg(MSG_USR_OUT, SEP);
	if (html) {
	    fprintf(stdout, "<p>\n");
	    fprintf(stdout, "<center>\n");
	    fprintf(stdout, "<table width=\"80%%\" border=3 cellpadding=2><tr><td>\n");
	    fprintf(stdout, "<table width=\"100%%\" border=1 cellpadding=2>\n");
	    dict_2html(dp, gif);
	    fprintf(stdout, "</table></td></tr>\n");
	    fprintf(stdout, "</table>\n");
	    fprintf(stdout, "</center>\n");
	} else
	    dict_print(dp, 1);
    }
#if SERVER
    if (!interactive && !html) (void) dict_fmt(LB_EOT, " ", 1);
#endif
#if 0
    if (html) {
	fprintf(stdout, "</BODY>\n");
	fprintf(stdout, "</HTML>\n");
    }
#endif
}

static void
dict_collect(d, p, f)
/*
 * Calculate the number of records and bytes emited
 * by the hits of "p" in dictionary "d"
 */
GODICT *d;		/* Dictionary		*/
char   *p;		/* search pattern	*/
FOUND  *f;		/* What did we find?	*/
{
    GODICT *dp;

    memset(f, 0, sizeof(FOUND));

    for (dp = dict_search(d, p, f); dp; dp = dict_search(dp->dct_next, p, f)) {
	f->fnd_records += 1;
	f->fnd_bytes   += dict_print(dp, 0);
	f->fnd_bytes   += strlen(SEP);		/* record seperator */
    }
#if SERVER
    /*
     * search seperator
     */
    if (!interactive) f->fnd_bytes += dict_fmt(LB_EOT," ",0);
#endif
}

void
dict_retrieve(d, p, f)
GODICT *d;
char   *p;		/* Pattern to look for	*/
FOUND  *f;		/* Result of the query	*/
{
    /*
     * Retrieve information from the database
     */
    dict_collect(d, p, f);

    /*
     * Output the retrieval information
     */
    if (MAXOUTPUT && f->fnd_bytes > MAXOUTPUT)	dict_report(f);
    else if (f->fnd_records == 0)		dict_report(f);
    else					dict_output(d, p, f);
}

void
dict_html(int mode)
{
    html = mode;
}

void
dict_interact(mode)
int mode;
{
    interactive = mode;
}

void
dict_verbose(mode)
int mode;
{
    verbose = mode;
}

void
dict_types(t)
/*
 * Select the types searched in the dictionary
 */
int t;
{
    types = t;
}

void
dict_langs(l)
/*
 * Select the languages returned from the dictionary
 */
int l;
{
    langs = l;
}

char *
lstr(s)
/*
 * Convert a string to lowercase
 */
char *s;
{
    char *t;

    for (t = s; *t; t++) if (isupper(*t)) *t = tolower(*t);

    return s;
}
